// Canon BubbleJet BJC series driver Class
//
// 2001, Philippe Houdoin
// Based on:
// - gdevbjc.c from Aladdin GhostScript
// - LMDriver.cpp from Ryan Lockhart Lexmark 5/7xxxx BeOS driver 
//
// Printers supported
//   bj30, bj100, bj200, bj200e, bj200ex, bj230, bjc70, bjc210,
//   bjc600, bjc600e, bjc800, bjc4000 and bjc4100
// 
// Todo list:
// - color support
// - add landscape
// - decrease memory needs: multiple partial rendition?
// - gamma (global or by color) correction?

#define DEBUG	1

#include <Debug.h>

#include <stdio.h>
#include <string.h>			// for memset()

#include <StorageKit.h>

#include "BJC.h"

#define ROUND_UP(x, y) (((x) + (y) - 1) & ~((y) - 1))

// Constructor & destructor
// ------------------------

BJC::BJC()
	: Printer()
{
	x_resolution 	= 360;	// 360 dpi
	y_resolution 	= 360;	// 360 dpi
	
	m_page_bitmap		= NULL;
	
	
}

BJC::~BJC()
{
	if ( m_page_bitmap )
	 	delete m_page_bitmap;
	
	if ( transport )
		{
		ExitPrinter();
		CloseTransport();
		};
}


// Methods
// -------

status_t BJC::PrintPage
	(
	BFile *		spool,
	BMessage *	job_msg,
	int32		page_number,
	int32		page_count
	)
{
	status_t	status;
	// char 		text[64];

	int *		errors;
	int *		first_line_errors;
	int * 		second_line_errors;
	int * 		current_error;

	uint8 * 	ptr;
	int32		bpr;
	int32		x, y, w, h;
	uint32		errors_row_width;
	rgb_color 	color;
	int 		gray;
	int8		error;

	int			skip;
	uint8 *		scan_line;
	

	status = RasterizePage(spool, job_msg);

	w 	= m_page_bitmap->Bounds().IntegerWidth();
	h	= m_page_bitmap->Bounds().IntegerHeight();
	bpr = m_page_bitmap->BytesPerRow(); 

	errors_row_width = w + 16; // 16 = dummy margins at left and right

	errors = (int *) malloc(2 * (sizeof(int) * errors_row_width) );

	first_line_errors = errors + 8;
	second_line_errors = errors + errors_row_width + 8;
	
	memset(first_line_errors, 0, sizeof(int) * errors_row_width );

	for ( y = 0; y < h - 1; y++ )
		{
		// For each row

		ptr = (uint8 *) m_page_bitmap->Bits();
		ptr += (y * bpr);

		current_error = first_line_errors;
		memset(second_line_errors, 0, sizeof(int) * errors_row_width );

		for ( x = 0; x < w - 1; x++ )
			{
			// For each pixel
			color.blue 	= *ptr;
			color.green	= *(ptr+1);
			color.red	= *(ptr+2);
			gray = (int) (0.30 * color.red + 0.59 * color.green + 0.11 * color.blue);
			gray += *current_error;
			
			if ( gray > 127 )
				{
				// mostly white
				*ptr 		= 255;	// blue
				*(ptr+1)	= 255;	// green
				*(ptr+2)	= 255;	// red
				error = gray - 255;
				}
			else
				{
				// mostly black
				*ptr 		= 0;	// blue
				*(ptr+1)	= 0;	// green
				*(ptr+2)	= 0;	// red
				error = gray;
				};

			// diffuse this gray error to sibblings pixels
			// Burkes dithering matrix :
			//
			//     x 8 4
			// 2 4 8 4 2

			*(current_error + 1) 			+= (8 * error) / 32;
			*(current_error + 2) 			+= (4 * error) / 32;

			*(second_line_errors + x - 2) 	+= (2 * error) / 32;
			*(second_line_errors + x - 1) 	+= (4 * error) / 32;
			*(second_line_errors + x) 		+= (8 * error) / 32;
			*(second_line_errors + x + 1) 	+= (4 * error) / 32;
			*(second_line_errors + x + 2) 	+= (2 * error) / 32;

			ptr += 4;
			current_error++;
			};
			
		// swap errors line, so second one become first one at next row
		current_error 		= first_line_errors;
		first_line_errors 	= second_line_errors;
		second_line_errors 	= current_error;
		};
		
	free(errors);
	
	// m_page_bitmap->Unlock();

/*
#ifdef DEBUG

	// Show bitmap in a window
	sprintf(text, "Printing page %ld/%ld...", page_number, page_count);
	BitmapWindow *	bw = new BitmapWindow(text, m_page_bitmap);
	bw->Go();
#endif
*/

	if ( page_number == 1 )
		// Init printer before first page
		InitPrinter();

	// Set data compression on ESC ( b 0x01 0x00 <bool>
	WriteCommand('b', 1, "\x01");

	BeginPage();

	// avance paper one inch 
	// AdvancePaper(y_resolution);

	// avance print head one inch to the right
	// AdvanceHead(x_resolution);
	
	scan_line = (uint8 *) malloc(w + (w % 128) + 1 + 16); // +16 = margin

	skip = 0;

	for ( y = 0; y < h - 1; y++ )
		{
		// For each scanline
		int count;

		ptr = (uint8 *) m_page_bitmap->Bits();
		ptr += (y * bpr);

		count = BuildScanLine(ptr, w, scan_line );
		if ( count )
			{
			if ( skip )
				AdvancePaper(skip);
			skip = 1;

			// Raster Plane start
			WriteCommand('A', count + 1);	// +1 = Ink code, first byte of data
			transport->Write("K", 1);	// Black ink
			transport->Write(scan_line, count);
			transport->Write("\x0d", 1);		// Raster Plane end

			}
		else
			skip++;

		}; // for each scanline

	if ( skip )
		AdvancePaper(skip);

	EndPage();

	free(scan_line);

	if ( m_page_bitmap )
		delete m_page_bitmap;
		
	m_page_bitmap = NULL;

	return status;
}


// ----------------------------
int BJC::BuildScanLine
	(
	const uint8 * 	bits,
	int				width,
	uint8 *			scan_line
	)
{
	const uint8 *	in;
	uint8 *			buffer;
	uint8 *			out;
	uint8			byte;
	int32			x;
	int 			bit;
	int 			count;
	bool			empty_line;

	bit = 7;
	byte = 0;
	
	empty_line = true;

	in 		= bits;
	out 	= scan_line;
	count 	= width;
	// nb bits by row to nb bytes per row
	if (count % 8)
		count = (count / 8) + 1;
	else
		count = count / 8;

	buffer = (uint8 *) malloc(count + 16);	// +16=margin
	out = buffer;

	for ( x = 0; x < width - 1; x++ )
		{
		if ( *in == 0 )	// red, green, blue are 0 (black) or 255 (white)
			{
			byte |= (1 << bit);
			empty_line = false;
			};
			
		in += 4;	// next pixel
		bit--;
		if ( bit < 0 )
			{
			*out = byte;
			out++;
			byte = 0;
			bit = 7;
			};
		}; // for each colum
		
	if ( bit >= 0 )
		{
		// write last byte...
		*out = byte;
		out++;
		};

	if ( empty_line )
		count = 0;
	else
		count = CompressFrame(buffer, count, scan_line);
	
	free(buffer);
	return count;
		
}

// --------------------------------------------------
int BJC::CompressFrame
	(
	const uint8 *	row,
	int 			raster,
	uint8 *			compressed
	)
{
  const uint8 * end_row = row;
  register const uint8 * exam = row;
  register uint8 * cptr = compressed; /* output pointer into compressed bytes */
  
  end_row += raster;

  while ( exam < end_row ) {
    /* Search ahead in the input looking for a run */
    /* of at least 4 identical bytes. */
    const uint8 *compr = exam;
    const uint8 *end_dis;
    const uint8 *next;
    register uint8 test, test2;

    test = *exam;
    while ( exam < end_row ) {
      test2 = *++exam;
      if ( test == test2 )
	  break;
      test = test2;
    }
    

    /* Find out how long the run is */
    end_dis = exam - 1;
    if ( exam == end_row ) { /* no run */
      next = --end_row;
    } else { 

      next = exam + 1;
      while ( next < end_row && *next == test ) next++;
    }
    

    /* Now [compr..end_dis) should be encoded as dissimilar, */
    /* and [end_dis..next) should be encoded as similar. */
    /* Note that either of these ranges may be empty. */
    

    for ( ; ; ) {	/* Encode up to 128 dissimilar bytes */
      uint count = end_dis - compr; /* uint for faster switch */
      switch ( count ) { /* Use memcpy only if it's worthwhile. */
      case 6: cptr[6] = compr[5];
      case 5: cptr[5] = compr[4];
      case 4: cptr[4] = compr[3];
      case 3: cptr[3] = compr[2];
      case 2: cptr[2] = compr[1];
      case 1: cptr[1] = compr[0];
	*cptr = count - 1;
	cptr += count + 1;
      case 0: /* all done */
	break;
      default:
	if ( count > 128 ) count = 128;
	*cptr++ = count - 1;
	memcpy(cptr, compr, count);
	cptr += count, compr += count;
	continue;
      }
      break;
    }
    

    {	/* Encode up to 128 similar bytes. */
      /* Note that count may be <0 at end of row. */
      int count = next - end_dis;
      if (next < end_row || test != 0)
	while ( count > 0 ) { 

	  int this_one = (count > 128 ? 128 : count);
	  *cptr++ = 257 - this_one;
	  *cptr++ = (uint8) test;
	  count -= this_one;
	}
      exam = next;
    }
  }
  return (int)(cptr - compressed);
}

			

status_t BJC::InitPrinter()
{
	// Reset to default setting...
	transport->Write("\x1b[K\x02\x00\x00\x0f", 7);
	
	// ???  ESC ( a <count lo> <count hi> ...
	WriteCommand('a', 1, "\x01");
	
	// Set media supply method ESC ( l <count lo> <count hi> <param1> <param2>
	WriteCommand('l', 1, "\x14");	// 14 = ????
	
	// Set page margins ESC ( g <count lo> <count hi> ...
	WriteCommand('g', 3, "\x75\x01\x72"); // = ????
	
	// Set raster resolution ESC ( d <count lo> <count hi> <y_res> [<x_res>]
	if ( x_resolution == y_resolution )
		WriteCommand('d', 2);
	else
		{
		WriteCommand('d', 4);
		WriteHiLo(y_resolution);
		}
	WriteHiLo(x_resolution);
	return B_OK;
}


status_t BJC::ExitPrinter()
{
	if (! transport)
		return B_ERROR;
		
	// ???  ESC ( a <count lo> <count hi> ...
	WriteCommand('a', 1, "\x00");
	
	// Set data compression off ESC ( b 0x01 0x00 <bool>
	WriteCommand('b', 1, "\x00");
	
	// Initial condition ESC @
	return transport->Write("\x1b@", 2);
}


status_t BJC::BeginPage()
{
	// Carriage Return...
	return transport->Write("\x0d", 1);
}



status_t BJC::EndPage()
{
	// Form Feed
	return transport->Write("\x0c", 1);
}


status_t BJC::AdvancePaper
	(
	int		count
	)
{
	WriteCommand('e', 2);
	WriteHiLo(count);

	return B_OK;
}


status_t BJC::AdvanceHead
	(
	int		count
	)
{
/*
	int 	nb;
	uint8	byte;

	// print head has only 120 dpi step precision

	nb	   = 120 * count / x_resolution;
	count -= (x_resolution * nb / 120); 

	// first, quick horizontal skip, at 120 dpi resolution
	transport->Write("\033d", 2);
	byte = nb & 0xFF;
	transport->Write(&byte, 1);
	byte = nb >> 8;
	transport->Write(&byte, 1);
	
	if ( count )
		{
		// second, print nothing up to x_resolution * count dots

		int i;

		count *= head_nb_bytes;	// 6*8 bits on print head (48 dots)

		transport->Write("\033[g", 3);
		byte = count & 0xFF;
		transport->Write(&byte, 1);
		byte = count >> 8;
		transport->Write(&byte, 1);
		byte = graphic_mode;
		transport->Write(&byte, 1);
		
		for ( i = 0; i < count ; i++)
			transport->Write("\000\000\000\000\000\000", 6);
		};
*/		
	return B_OK;
}


status_t BJC::RasterizePage
	(
	BFile *		spool,
	BMessage *	msg
	)
{
	BView * view;
	long picture_count;
	BRect print_rect;
	BRect * pic_rects;
	BPoint * pic_points;
	BRegion * pic_region;
	BPicture ** pictures;
	// BAlert * alert;
	// char text[128];
	
	int32 pic;

	print_rect = msg->FindRect("printable_rect");
	
	spool->Read(&picture_count, sizeof(long));

	// sprintf(text, "picture_count = %ld", picture_count);
	// alert = new BAlert("RasterizePage", text, "Ok");
	// alert->Go();

		
	pictures = (BPicture **) 	malloc(picture_count * sizeof(BPicture *));
	pic_rects = (BRect *)		malloc(picture_count * sizeof(BRect));
	pic_points = (BPoint *)		malloc(picture_count * sizeof(BPoint));
	pic_region = new BRegion();
	
	// alert = new BAlert("RasterizePage", "BPictures allocated", "Ok");
	// alert->Go();

	for (pic = 0; pic < picture_count; pic++)
	{
		spool->Seek(40 + sizeof(off_t), SEEK_CUR);
		spool->Read(&pic_points[pic], sizeof(BPoint));
		spool->Read(&pic_rects[pic], sizeof(BRect));
		pictures[pic] = new BPicture();
		pictures[pic]->Unflatten(spool);
		pic_region->Include(pic_rects[pic]);
	}

	// alert = new BAlert("RasterizePage", "BPictures loaded", "Ok");
	// alert->Go();

	float scale = x_resolution / 72.0;

	BRect bitmap_rect(0, 0, (pic_region->Frame().right - pic_region->Frame().left) * scale, 
						    (pic_region->Frame().bottom - pic_region->Frame().top) * scale);	
	
	view = new BView(bitmap_rect, "", B_FOLLOW_ALL, B_WILL_DRAW);
	m_page_bitmap = new BBitmap(bitmap_rect, B_RGB32, true, false); // B_GRAY8, true, false);

	// sprintf(text, "BBitmap(%f, %f, %f, %f) created", bitmap_rect.left, bitmap_rect.top,
	// 		bitmap_rect.right, bitmap_rect.bottom);
	// alert = new BAlert("RasterizePage", text, "Ok");
	// alert->Go();

	m_page_bitmap->AddChild(view);
	
	view->Window()->Lock();
	view->SetScale(scale);
	
	// Okay, white background...
	view->SetHighColor(255,255,255);
	view->FillRect(view->Bounds());

	view->MovePenTo(0,0);
		
	// alert = new BAlert("RasterizePage", "Ready to draw BPictures", "Ok");
	// alert->Go();

	for (pic = 0; pic < picture_count; pic++)
		{
		// view->MovePenBy(0, vertical_offset);
		view->DrawPicture(pictures[pic]);
		};
		
	// alert = new BAlert("RasterizePage", "BPictures drawn", "Ok");
	// alert->Go();

	view->Sync();

	view->RemoveSelf();
	
	delete pic_region;

	free(pic_points);
	free(pic_rects);
	
	for (pic = 0; pic < picture_count; pic++)
		delete pictures[pic];
	free(pictures);
	
	return B_OK;
}


status_t BJC::WriteCommand
	(
	uint8 			command,
	int				length = 0,
	const void *	data = NULL
	)
{
	uint8	tmp[3] = { 0x1b, '(', ' ' };
	
	tmp[2] = command;
	
	transport->Write(tmp, 3);
	WriteLoHi(length);
	
	if ( data != NULL && length > 0 )
		transport->Write(data, length);
		
	return B_OK;
}


status_t BJC::WriteLoHi(int value)
{
	uint8 tmp[2];
	
	tmp[0] = (value & 0xff);
	tmp[1] = (value & 0xffff) >> 8;
	
	return transport->Write(tmp, 2);
}

	 
status_t BJC::WriteHiLo(int value)
{
	uint8 tmp[2];
	
	tmp[0] = (value & 0xffff) >> 8;
	tmp[1] = (value & 0xff);
	
	return transport->Write(tmp, 2);
}
 
