// Printer driver genberic class
//
// 2001, Philippe Houdoin
//

#define DEBUG	1

#include <Debug.h>

#include <stdio.h>
#include <StorageKit.h>

#include "Printer.h"
#include "PrinterSetupWindow.h"
#include "PageSetupWindow.h"
#include "JobSetupWindow.h"



// Constructor & destructor
// ------------------------

Printer::Printer()
{
	transport_add_on = 0;
	transport = NULL;
}


Printer::~Printer()
{
}


// Methods
// -------

status_t Printer::PrintJob
	(
	BFile *		spool,		// spool file
	BNode * 	printer,	// printer node, used by OpenTransport() to find & load transport add-on
	BMessage * 	msg			// job message
	)
{
	print_file_header	pfh;
	status_t			status;
	BMessage * 			job_msg;
	int32 				page;
	uint32				copy;
	uint32				copies;

	// open transport
	if ( OpenTransport(printer) != B_OK )
		return B_ERROR;

	// read print file header	
	spool->Seek(0, SEEK_SET);
	spool->Read(&pfh, sizeof(pfh));
	
	// read job message
	job_msg = new BMessage();
	job_msg->Unflatten(spool);
	
	if ( job_msg->HasInt32("copies") )
		copies = job_msg->FindInt32("copies");
	else
		copies = 1;
	
	status = B_OK;
	for ( copy = 0; copy < copies; copy++ )
		{	
		for ( page = 1; page <= pfh.page_count && status == B_OK; page++ )
			status = PrintPage(spool, job_msg, page, pfh.page_count);

		if ( status != B_OK )
			break;

		// re-read job message for next page
		spool->Seek(sizeof(pfh), SEEK_SET);
		job_msg->Unflatten(spool);
		};
		
	CloseTransport();

	return status;
}


status_t Printer::PrintPage
	(
	BFile *		spool,
	BMessage *	job_msg,
	int32		page_number,
	int32		page_count
	)
{
#ifdef DEBUG
	char 		text[128];

	sprintf(text, "Faking print of page %ld/%ld...", page_number, page_count);
	BAlert *	alert = new BAlert("Printer::PrintPage()", text, "Hmm?");
	alert->Go();
	return B_OK;
#else
	return B_UNSUPPORTED;
#endif
}



status_t Printer::PrinterSetup
	(
	char * 	printer_name	// name of printer, to attach printer settings
	)
{
	PrinterSetupWindow * psw;

	psw = new PrinterSetupWindow(printer_name);
	return psw->Go();
}


status_t Printer::PageSetup
	(
	BMessage *	msg			// page setup message
	)
{
	BRect paper_rect;
	BRect print_rect;

#define SCALE 8.3333f
#define SCREEN 72.0f
#define LETTER_WIDTH 8.5
#define LETTER_HEIGHT 11

	if (msg->HasInt64("xres"))
		msg->ReplaceInt64("xres", 360);
	else
		msg->AddInt64("xres", 360);
		
	if (msg->HasInt64("yres"))
		msg->ReplaceInt64("yres", 360);
	else
		msg->AddInt64("yres", 360);
				
	if (msg->HasInt32("orientation"))
		msg->ReplaceInt32("orientation", 1);
	else
		msg->AddInt32("orientation", 1);
			
	paper_rect = BRect(0, 0, LETTER_WIDTH * SCREEN, LETTER_HEIGHT * SCREEN);
	print_rect = BRect(0, 0, LETTER_WIDTH * SCREEN, LETTER_HEIGHT * SCREEN);
 	
	if (msg->HasRect("paper_rect"))
		msg->ReplaceRect("paper_rect", paper_rect);
	else
		msg->AddRect("paper_rect", paper_rect);

	if (msg->HasRect("printable_rect"))
		msg->ReplaceRect("printable_rect", print_rect);
	else
		msg->AddRect("printable_rect", print_rect);
	 		
	if (msg->HasFloat("scaling"))
		msg->ReplaceFloat("scaling", SCREEN);
	else
		msg->AddFloat("scaling", SCREEN);

	PageSetupWindow * psw;
	
	psw = new PageSetupWindow(msg);
	return psw->Go();
}



status_t Printer::JobSetup
	(
	BMessage *	msg				// job setup message
	)
{
	if (msg->HasInt32("copies"))
		msg->ReplaceInt32("copies", 1);
	else
		msg->AddInt32("copies", 1);

	if (msg->HasInt32("first_page"))
		msg->ReplaceInt32("first_page", 1);
	else
		msg->AddInt32("first_page", 1);
		
	if (msg->HasInt32("last_page"))
		msg->ReplaceInt32("last_page", 0xffff);
	else
		msg->AddInt32("last_page", 0xffff);

	JobSetupWindow * jsw;

	jsw = new JobSetupWindow(msg);
	return jsw->Go();
}


status_t Printer::OpenTransport
	(
	BNode * 	printer			// node of printer contains transport add-on name
	)
{
	char 		buffer[512];
	BPath * 	path;
	
	// first, find & load transport add-on
	path = new BPath();
	
	// find name of this printer transport add-on 
	printer->ReadAttr("transport", B_STRING_TYPE, 0, buffer, sizeof(buffer));
	
	// try first on user add-ons directory
	find_directory(B_USER_ADDONS_DIRECTORY, path);
	path->Append("Print/transport");
	path->Append(buffer);
	transport_add_on = load_add_on(path->Path());
	
	if (transport_add_on < 0)
		{
		// add-on not in user add-ons directory. try system one
		find_directory(B_BEOS_ADDONS_DIRECTORY, path);
		path->Append("Print/transport");
		path->Append(buffer);
		transport_add_on = load_add_on(path->Path());
		};
	
	if (transport_add_on < 0)
		{
		BAlert * alert = new BAlert("Uh oh!", "Couldn't find transport add-on.", "OK");
		alert->Go();
		return B_ERROR;
		};
	
	// get init & exit proc
	get_image_symbol(transport_add_on, "init_transport", B_SYMBOL_TYPE_TEXT, (void **) &init_transport);
	get_image_symbol(transport_add_on, "exit_transport", B_SYMBOL_TYPE_TEXT, (void **) &exit_transport);
	
	if ( ! init_transport ||
		 ! exit_transport )
		{
		BAlert * alert = new BAlert("Uh oh!", "Couldn't resolve transport symbols.", "OK");
		alert->Go();
		return B_ERROR;
		};
	
	delete path;
	
	// now, init transport add-on
	node_ref 	ref;
	BDirectory 	dir;

	printer->GetNodeRef(&ref);
	dir.SetTo(&ref);
	
	path = new BPath(&dir, NULL);
	strcpy(buffer, path->Path());
	
	// create BMessage for init_transport()
	BMessage *msg = new BMessage('TRIN');
	msg->AddString("printer_file", buffer);
	
	transport = (*init_transport)(msg);
	
	delete msg;
	delete path;
	
	if (transport == 0)
		{
		BAlert * alert = new BAlert("Uh oh!", "Couldn't open transport.", "OK");
		alert->Go();
		return B_ERROR;
		};
	
	return B_OK;
}


status_t Printer::CloseTransport()
{
	if ( ! transport_add_on )
		return B_ERROR;

	if ( exit_transport )
		(*exit_transport)();

	unload_add_on(transport_add_on);
	transport_add_on = 0;
	transport = NULL;
	
	return B_OK;
}

